treechildren::-moz-tree-checkbox { /* css for unchecked cells */
  list-style-image: url("chrome://https-everywhere/skin/cross.png");
}

treechildren::-moz-tree-checkbox(checked) { /* css for checked cells */
  list-style-image: url("chrome://https-everywhere/skin/tick.png");
}

treechildren::-moz-tree-checkbox(undefined) { /* css for empty cells */
  list-style-image: url("");
}
